#include <stdio.h>
#include <stdlib.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>


#include "matrix.h"
#include "display.h"
#include "object.h"

int             nvertex, npoly;
point          *vertex_list;
vector         *new_vertex_list;
int           **poly_list;

#define VERTEX(n,x) vertex_list[ poly_list[ n][ x]]
#define MIN(x,y) ((x)<(y)?x:y)
#define MAX(x,y) ((x)>(y)?x:y)

void            init_object(datafile, xs, ys, zs)
char           *datafile;
float           xs, ys, zs;
{
   FILE           *fptr;
   float           minx, maxx, miny, maxy, minz, maxz;
   float           xx, yy, zz;
   int             i;

   if (!(fptr = fopen(datafile, "r")))
   {
      fprintf(stderr, "Error opening %s\n", datafile);
      exit(1);
   }

   minx = miny = minz = 999999.0;
   maxx = maxy = maxz = 0.0;

   fscanf(fptr, "%d", &nvertex);
   vertex_list = (point *) malloc(sizeof(point) * nvertex);
   new_vertex_list = (vector *) malloc(sizeof(vector) * nvertex);

   for (i = 0; i < nvertex; i++)
   {
      /*
       * switch y's and z's.
       */
      fscanf(fptr, "%f %f %f", &xx, &yy, &zz);
      xx *= xs;
      yy *= ys;
      zz *= zs;

      vertex_list[i].x = (double) xx;
      vertex_list[i].y = (double) yy;
      vertex_list[i].z = (double) zz;
      minx = MIN(xx, minx);
      miny = MIN(yy, miny);
      minz = MIN(zz, minz);

      maxx = MAX(xx, maxx);
      maxy = MAX(yy, maxy);
      maxz = MAX(zz, maxz);
   }
   fscanf(fptr, "%d", &npoly);
   poly_list = (int **) malloc(sizeof(int *) * npoly);
   for (i = 0; i < npoly; i++)
   {
      poly_list[i] = (int *) malloc(sizeof(int) * 3);
      fscanf(fptr, "%d %d %d",
         &(poly_list[i][0]),
         &(poly_list[i][1]),
         &(poly_list[i][2]));
   }

   /*
    * printf( "done reading: # vertices= %d, # polys = %d \n",
    * nvertex,npoly);
    */
   printf("minimum: <%f %f %f>\n", minx, miny, minz);
   printf("maximum: <%f %f %f>\n", maxx, maxy, maxz);
#if 1
   for (i = 0; i < nvertex; i++)
   {
      vertex_list[i].x -= minx;
      vertex_list[i].y -= miny;
      vertex_list[i].z -= minz;
   }
#endif
}

void transform_object()
{
   extern matrix T_matrix;

   double *Matrix_row_0 = T_matrix[0];
   double *Matrix_row_1 = T_matrix[1];
   double *Matrix_row_2 = T_matrix[2];
   double *Matrix_row_3 = T_matrix[3];

   double x_hat;
   double y_hat;
   double z_hat;

   int  i;

   for (i = 0; i < nvertex; i++)
   {
      x_hat = vertex_list[i].x;
      y_hat = vertex_list[i].y;
      z_hat = vertex_list[i].z;

      new_vertex_list[i][0] = Matrix_row_0[0] * x_hat + Matrix_row_0[1] * y_hat
                              + Matrix_row_0[2] * z_hat + Matrix_row_0[3];

      new_vertex_list[i][1] = Matrix_row_1[0] * x_hat + Matrix_row_1[1] * y_hat
                             + Matrix_row_1[2] * z_hat + Matrix_row_1[3];

      new_vertex_list[i][2] = Matrix_row_2[0] * x_hat + Matrix_row_2[1] * y_hat
                              + Matrix_row_2[2] * z_hat + Matrix_row_2[3];

      new_vertex_list[i][3] = Matrix_row_3[0] * x_hat + Matrix_row_3[1] * y_hat
                             + Matrix_row_3[2] * z_hat + Matrix_row_3[3];
   }
}

void            draw_object()
{
   extern Display *theDisplay;
   extern GC       theGC;
   extern long     theForegroundPixel;
   extern long     theBackgroundPixel;
   extern int      height;
   extern int      width;
   extern Pixmap   thePixmap;
   extern Window   theDA;

   register int    i;

   int    v0;
   int    v1;
   int    v2;

   XSetForeground(theDisplay, theGC, theBackgroundPixel);
   XFillRectangle(theDisplay, thePixmap, theGC, 0, 0, width, height);

   for (i = 0; i < npoly; i++)
   {
#ifndef TRANSFORM
      draw_line(&VERTEX(i, 0), &VERTEX(i, 1), 0);
      draw_line(&VERTEX(i, 1), &VERTEX(i, 2), 0);
      draw_line(&VERTEX(i, 2), &VERTEX(i, 0), 0);

#else
      v0 = poly_list[i][0];
      v1 = poly_list[i][1];
      v2 = poly_list[i][2];

      draw_line(v0, v1);
      draw_line(v1, v2);
      draw_line(v2, v0);
#endif
   }
   XCopyArea(theDisplay, thePixmap, theDA, theGC, 0, 0, width, height, 0, 0);
}

#ifndef TRANSFORM
void            draw_line(p1, p2, col)
point          *p1, *p2;
int             col;
{
   extern Display  *theDisplay;
   extern Pixmap    thePixmap;
   extern GC        theGC;
   extern long      theForegroundPixel;
   vector          v1, v2;

   XSetForeground(theDisplay, theGC, theForegroundPixel);
   point2vector(v1, p1);
   point2vector(v2, p2);
   if (pipe_vector(v1, v2))
   {
      XDrawLine(theDisplay, thePixmap, theGC,
         (int) v1[0], (int) v1[1], (int) v2[0], (int) v2[1]);
   }
}
#else
void draw_line(int v0, int v1)
{
   extern Display  *theDisplay;
   extern Pixmap    thePixmap;
   extern GC        theGC;
   extern long      theForegroundPixel;

   extern vector *new_vertex_list;
   extern double  Zmin;
   extern matrix  vp_matrix;

   double x0;
   double y0;
   double z0;
   double h0;
   double x1;
   double y1;
   double z1;
   double h1;

   XSetForeground(theDisplay, theGC, theForegroundPixel);

   x0 = new_vertex_list[v0][0];
   y0 = new_vertex_list[v0][1];
   z0 = new_vertex_list[v0][2];
   h0 = new_vertex_list[v0][3];

   x1 = new_vertex_list[v1][0];
   y1 = new_vertex_list[v1][1];
   z1 = new_vertex_list[v1][2];
   h1 = new_vertex_list[v1][3];

   if (clip_3d(&x0, &y0, &z0, &x1, &y1, &z1, Zmin))
   {
#if 0
      printf("line form (%g, %g, %g) to (%g, %g, %g)\n", x0, y0, z0,
             x1, y1, z1);
#endif
      homogenize_vector_transform(&x0, &y0, &z0, &h0, vp_matrix);
      homogenize_vector_transform(&x1, &y1, &z1, &h0, vp_matrix);
      XDrawLine(theDisplay, thePixmap, theGC,
                (int) x0, (int) y0, (int) x1, (int) y1);

   }
}

#endif
